# Generated by Django 5.1.1 on 2024-11-22 16:32

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_tasks_database", "0011_rename_complete_status"),
    ]

    operations = [
        migrations.AddField(
            model_name="dbtaskresult",
            name="exception_class_path",
            field=models.TextField(default="", verbose_name="exception class path"),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="dbtaskresult",
            name="traceback",
            field=models.TextField(default="", verbose_name="traceback"),
            preserve_default=False,
        ),
    ]



from django.core.exceptions import ImproperlyConfigured


class TaskException(Exception):  # noqa: N818
    """Base class for task-related exceptions. Do not raise directly."""


class InvalidTaskError(TaskException):
    """
    The provided Task is invalid.
    """


class InvalidTaskBackendError(ImproperlyConfigured):
    pass


class TaskResultDoesNotExist(TaskException):
    pass


class TaskResultMismatch(TaskException):
    pass



# Generated by Django 5.1.1 on 2024-11-22 16:32

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("django_tasks_database", "0013_separate_exception_fields"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="dbtaskresult",
            name="exception_data",
        ),
    ]



# Generated by Django 5.0.6 on 2024-06-13 10:24

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_tasks_database", "0002_alter_dbtaskresult_options"),
    ]

    operations = [
        migrations.AddField(
            model_name="dbtaskresult",
            name="enqueued_at",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="dbtaskresult",
            name="finished_at",
            field=models.DateTimeField(null=True),
        ),
    ]



import datetime
import logging
import uuid
from typing import TYPE_CHECKING, Any, Generic, Optional, TypeVar

import django
from django.conf import settings
from django.core.exceptions import SuspiciousOperation
from django.db import models
from django.db.models import F, Q
from django.db.models.constraints import CheckConstraint
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from typing_extensions import ParamSpec

from django_tasks.base import (
    DEFAULT_TASK_QUEUE_NAME,
    TASK_DEFAULT_PRIORITY,
    TASK_MAX_PRIORITY,
    TASK_MIN_PRIORITY,
    Task,
    TaskError,
    TaskResultStatus,
)
from django_tasks.compat import TASK_CLASSES
from django_tasks.utils import get_exception_traceback, get_module_path, retry

from .utils import normalize_uuid

logger = logging.getLogger("django_tasks.backends.database")

T = TypeVar("T")
P = ParamSpec("P")

if TYPE_CHECKING:
    from .backend import TaskResult

    class GenericBase(Generic[P, T]):
        pass

else:

    class GenericBase:
        """
        https://code.djangoproject.com/ticket/33174
        """

        def __class_getitem__(cls, _):
            return cls


def get_date_max() -> datetime.datetime:
    return datetime.datetime(
        9999, 1, 1, tzinfo=datetime.timezone.utc if settings.USE_TZ else None
    )


class DBTaskResultQuerySet(models.QuerySet):
    def ready(self) -> "DBTaskResultQuerySet":
        """
        Return tasks which are ready to be processed.
        """
        return self.filter(
            status=TaskResultStatus.READY,
        ).filter(
            models.Q(run_after=get_date_max()) | models.Q(run_after__lte=timezone.now())
        )

    def succeeded(self) -> "DBTaskResultQuerySet":
        return self.filter(status=TaskResultStatus.SUCCEEDED)

    def failed(self) -> "DBTaskResultQuerySet":
        return self.filter(status=TaskResultStatus.FAILED)

    def running(self) -> "DBTaskResultQuerySet":
        return self.filter(status=TaskResultStatus.RUNNING)

    def finished(self) -> "DBTaskResultQuerySet":
        return self.failed() | self.succeeded()

    @retry()
    def get_locked(self) -> Optional["DBTaskResult"]:
        """
        Get a job, locking the row and accounting for deadlocks.
        """
        return self.select_for_update(skip_locked=True).first()


class DBTaskResult(GenericBase[P, T], models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    status = models.CharField(
        _("status"),
        choices=TaskResultStatus.choices,
        default=TaskResultStatus.READY,
        max_length=max(len(value) for value in TaskResultStatus.values),
    )

    enqueued_at = models.DateTimeField(_("enqueued at"), auto_now_add=True)
    started_at = models.DateTimeField(_("started at"), null=True)
    finished_at = models.DateTimeField(_("finished at"), null=True)

    args_kwargs = models.JSONField(_("args kwargs"))

    priority = models.IntegerField(_("priority"), default=TASK_DEFAULT_PRIORITY)

    task_path = models.TextField(_("task path"))
    worker_ids = models.JSONField(_("worker id"), default=list)

    queue_name = models.CharField(
        _("queue name"), default=DEFAULT_TASK_QUEUE_NAME, max_length=32
    )
    backend_name = models.CharField(_("backend name"), max_length=32)

    run_after = models.DateTimeField(_("run after"))

    return_value = models.JSONField(_("return value"), default=None, null=True)

    exception_class_path = models.TextField(_("exception class path"))
    traceback = models.TextField(_("traceback"))

    objects = DBTaskResultQuerySet.as_manager()

    class Meta:
        ordering = [F("priority").desc(), F("run_after").asc()]
        verbose_name = _("Task Result")
        verbose_name_plural = _("Task Results")
        indexes = [
            models.Index(
                "status",
                *ordering,
                name="django_task_new_ordering_idx",
                condition=Q(status=TaskResultStatus.READY),
            ),
            models.Index(fields=["queue_name"]),
            models.Index(fields=["backend_name"]),
        ]

        if django.VERSION >= (5, 1):
            constraints = [
                CheckConstraint(
                    condition=Q(priority__range=(TASK_MIN_PRIORITY, TASK_MAX_PRIORITY)),
                    name="priority_range",
                )
            ]
        else:
            constraints = [
                CheckConstraint(
                    check=Q(priority__range=(TASK_MIN_PRIORITY, TASK_MAX_PRIORITY)),
                    name="priority_range",
                )
            ]

    @property
    def task(self) -> Task[P, T]:
        task = import_string(self.task_path)

        if not isinstance(task, TASK_CLASSES):
            raise SuspiciousOperation(
                f"Task {self.id} does not point to a Task ({self.task_path})"
            )

        return task.using(  # type: ignore[no-any-return]
            priority=self.priority,
            queue_name=self.queue_name,
            run_after=None if self.run_after == get_date_max() else self.run_after,
            backend=self.backend_name,
        )

    @property
    def task_result(self) -> "TaskResult[T]":
        from .backend import TaskResult

        task_result = TaskResult[T](
            db_result=self,
            task=self.task,
            id=normalize_uuid(self.id),
            status=TaskResultStatus[self.status],
            enqueued_at=self.enqueued_at,
            started_at=self.started_at,
            last_attempted_at=self.started_at,
            finished_at=self.finished_at,
            args=self.args_kwargs["args"],
            kwargs=self.args_kwargs["kwargs"],
            backend=self.backend_name,
            errors=[],
            worker_ids=self.worker_ids,
        )

        if self.status == TaskResultStatus.FAILED:
            task_result.errors.append(
                TaskError(
                    exception_class_path=self.exception_class_path,
                    traceback=self.traceback,
                )
            )

        object.__setattr__(task_result, "_return_value", self.return_value)

        return task_result

    @property
    def task_name(self) -> str:
        # If the function for an existing task is no longer available, it'll either raise an
        # ImportError or ModuleNotFoundError (a subclass of ImportError).
        try:
            return self.task.name
        except ImportError:
            pass

        try:
            return self.task_path.rsplit(".", 1)[1]
        except IndexError:
            return self.task_path

    @retry(backoff_delay=0)
    def claim(self, worker_id: str) -> None:
        """
        Mark as job as being run
        """
        self.status = TaskResultStatus.RUNNING
        self.started_at = timezone.now()
        self.worker_ids = [*self.worker_ids, worker_id]
        self.save(update_fields=["status", "started_at", "worker_ids"])

    @retry()
    def set_succeeded(self, return_value: Any) -> None:
        self.status = TaskResultStatus.SUCCEEDED
        self.finished_at = timezone.now()
        self.return_value = return_value
        self.exception_class_path = ""
        self.traceback = ""
        self.save(
            update_fields=[
                "status",
                "return_value",
                "finished_at",
                "exception_class_path",
                "traceback",
            ]
        )

    @retry()
    def set_failed(self, exc: BaseException) -> None:
        self.status = TaskResultStatus.FAILED
        self.finished_at = timezone.now()
        self.exception_class_path = get_module_path(type(exc))
        self.traceback = get_exception_traceback(exc)
        self.return_value = None
        self.save(
            update_fields=[
                "status",
                "return_value",
                "finished_at",
                "exception_class_path",
                "traceback",
            ]
        )



# Generated by Django 4.2.13 on 2024-06-08 09:10

import uuid

from django.db import migrations, models

import django_tasks.backends.database.models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="DBTaskResult",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("NEW", "New"),
                            ("RUNNING", "Running"),
                            ("FAILED", "Failed"),
                            ("COMPLETE", "Complete"),
                        ],
                        default="NEW",
                        max_length=8,
                    ),
                ),
                ("args_kwargs", models.JSONField()),
                ("priority", models.PositiveSmallIntegerField(default=0)),
                ("task_path", models.TextField()),
                ("queue_name", models.TextField(default="default")),
                ("backend_name", models.TextField()),
                ("run_after", models.DateTimeField(null=True)),
                ("result", models.JSONField(default=None, null=True)),
            ],
            options={
                "ordering": [
                    models.OrderBy(models.F("priority"), descending=True),
                    models.OrderBy(
                        models.F("run_after"), descending=True, nulls_last=True
                    ),
                ],
            },
            bases=(django_tasks.backends.database.models.GenericBase, models.Model),
        ),
    ]



# Generated by Django 4.2.13 on 2024-07-10 15:48

from django import VERSION
from django.db import migrations, models

if VERSION >= (5, 1):
    constraint = models.CheckConstraint(
        condition=models.Q(("priority__range", (-100, 100))), name="priority_range"
    )
else:
    constraint = models.CheckConstraint(
        check=models.Q(("priority__range", (-100, 100))), name="priority_range"
    )


class Migration(migrations.Migration):
    dependencies = [
        ("django_tasks_database", "0004_dbtaskresult_started_at"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dbtaskresult",
            name="priority",
            field=models.IntegerField(default=0),
        ),
        migrations.AddConstraint(
            model_name="dbtaskresult",
            constraint=constraint,
        ),
    ]



from collections.abc import Callable
from dataclasses import dataclass, field, replace
from datetime import datetime
from inspect import isclass, iscoroutinefunction
from typing import (
    TYPE_CHECKING,
    Any,
    Concatenate,
    Generic,
    Literal,
    TypeVar,
    cast,
    overload,
)

from asgiref.sync import async_to_sync, sync_to_async
from django.db.models.enums import TextChoices
from django.utils.module_loading import import_string
from django.utils.translation import pgettext_lazy
from django.utils.version import PY311
from typing_extensions import ParamSpec, Self

from .exceptions import TaskResultMismatch
from .utils import (
    get_module_path,
    normalize_json,
)

if TYPE_CHECKING:
    from .backends.base import BaseTaskBackend

DEFAULT_TASK_BACKEND_ALIAS = "default"
DEFAULT_TASK_QUEUE_NAME = "default"
TASK_MIN_PRIORITY = -100
TASK_MAX_PRIORITY = 100
TASK_DEFAULT_PRIORITY = 0

TASK_REFRESH_ATTRS = {
    "errors",
    "_return_value",
    "finished_at",
    "started_at",
    "last_attempted_at",
    "status",
    "enqueued_at",
    "worker_ids",
}


class TaskResultStatus(TextChoices):
    READY = ("READY", pgettext_lazy("Task", "Ready"))
    """The Task has just been enqueued, or is ready to be executed again."""

    RUNNING = ("RUNNING", pgettext_lazy("Task", "Running"))
    """The Task is currently running."""

    FAILED = ("FAILED", pgettext_lazy("Task", "Failed"))
    """The Task raised an exception during execution, or was unable to start."""

    SUCCEEDED = ("SUCCEEDED", pgettext_lazy("Task", "Succeeded"))
    """The Task has finished running successfully."""


T = TypeVar("T")
P = ParamSpec("P")


@dataclass(frozen=True, slots=PY311, kw_only=True)  # type: ignore[literal-required]
class Task(Generic[P, T]):
    priority: int
    """The Task's priority"""

    func: Callable[P, T]
    """The Task function"""

    backend: str
    """The name of the backend the Task will run on"""

    queue_name: str
    """The name of the queue the Task will run on"""

    run_after: datetime | None
    """The earliest this Task will run"""

    enqueue_on_commit: bool | None
    """
    Whether the Task will be enqueued when the current transaction commits,
    immediately, or whatever the backend decides.
    """

    takes_context: bool
    """
    Whether the Task receives the Task context when executed.
    """

    def __post_init__(self) -> None:
        self.get_backend().validate_task(self)

    @property
    def name(self) -> str:
        """
        An identifier for the task
        """
        return self.func.__name__

    def using(
        self,
        *,
        priority: int | None = None,
        queue_name: str | None = None,
        run_after: datetime | None = None,
        backend: str | None = None,
    ) -> Self:
        """
        Create a new Task with modified defaults.
        """

        changes: dict[str, Any] = {}

        if priority is not None:
            changes["priority"] = priority
        if queue_name is not None:
            changes["queue_name"] = queue_name
        if run_after is not None:
            changes["run_after"] = run_after
        if backend is not None:
            changes["backend"] = backend

        return replace(self, **changes)

    def enqueue(self, *args: P.args, **kwargs: P.kwargs) -> "TaskResult[T]":
        """
        Queue up the Task to be executed.
        """
        return self.get_backend().enqueue(self, args, kwargs)

    async def aenqueue(self, *args: P.args, **kwargs: P.kwargs) -> "TaskResult[T]":
        """
        Queue up the Task to be executed.
        """
        return await self.get_backend().aenqueue(self, args, kwargs)

    def get_result(self, result_id: str) -> "TaskResult[T]":
        """
        Retrieve the result for a Task of this type by id if it exists,
        otherwise raise ResultDoesNotExist.
        """
        result = self.get_backend().get_result(result_id)

        if result.task.func != self.func:
            raise TaskResultMismatch(
                f"Task does not match (received {result.task.module_path!r})"
            )

        return result

    async def aget_result(self, result_id: str) -> "TaskResult[T]":
        """See get_result()."""
        result = await self.get_backend().aget_result(result_id)

        if result.task.func != self.func:
            raise TaskResultMismatch(
                f"Task does not match (received {result.task.module_path!r})"
            )

        return result

    def call(self, *args: P.args, **kwargs: P.kwargs) -> T:
        if iscoroutinefunction(self.func):
            return async_to_sync(self.func)(*args, **kwargs)  # type:ignore[no-any-return]
        return self.func(*args, **kwargs)

    async def acall(self, *args: P.args, **kwargs: P.kwargs) -> T:
        if iscoroutinefunction(self.func):
            return await self.func(*args, **kwargs)  # type:ignore[no-any-return]
        return await sync_to_async(self.func)(*args, **kwargs)

    def get_backend(self) -> "BaseTaskBackend":
        from . import task_backends

        return task_backends[self.backend]

    @property
    def module_path(self) -> str:
        return get_module_path(self.func)


# Bare decorator usage
# e.g. @task
@overload
def task(function: Callable[P, T], /) -> Task[P, T]: ...


# Decorator with arguments
# e.g. @task() or @task(priority=1, ...)
@overload
def task(
    *,
    priority: int = TASK_DEFAULT_PRIORITY,
    queue_name: str = DEFAULT_TASK_QUEUE_NAME,
    backend: str = DEFAULT_TASK_BACKEND_ALIAS,
    enqueue_on_commit: bool | None = None,
    takes_context: Literal[False] = False,
) -> Callable[[Callable[P, T]], Task[P, T]]: ...


# Decorator with context and arguments
# e.g. @task(takes_context=True, ...)
@overload
def task(
    *,
    priority: int = TASK_DEFAULT_PRIORITY,
    queue_name: str = DEFAULT_TASK_QUEUE_NAME,
    backend: str = DEFAULT_TASK_BACKEND_ALIAS,
    enqueue_on_commit: bool | None = None,
    takes_context: Literal[True],
) -> Callable[[Callable[Concatenate["TaskContext", P], T]], Task[P, T]]: ...


# Implementation
def task(  # type: ignore[misc]
    function: Callable[P, T] | None = None,
    *,
    priority: int = TASK_DEFAULT_PRIORITY,
    queue_name: str = DEFAULT_TASK_QUEUE_NAME,
    backend: str = DEFAULT_TASK_BACKEND_ALIAS,
    enqueue_on_commit: bool | None = None,
    takes_context: bool = False,
) -> (
    Task[P, T]
    | Callable[[Callable[P, T]], Task[P, T]]
    | Callable[[Callable[Concatenate["TaskContext", P], T]], Task[P, T]]
):
    """
    A decorator used to create a task.
    """
    from . import task_backends

    def wrapper(f: Callable[P, T]) -> Task[P, T]:
        return task_backends[backend].task_class(
            priority=priority,
            func=f,
            queue_name=queue_name,
            backend=backend,
            enqueue_on_commit=enqueue_on_commit,
            takes_context=takes_context,
            run_after=None,
        )

    if function:
        return wrapper(function)

    return wrapper


@dataclass(frozen=True, slots=PY311, kw_only=True)  # type: ignore[literal-required]
class TaskError:
    exception_class_path: str
    traceback: str

    @property
    def exception_class(self) -> type[BaseException]:
        # Lazy resolve the exception class
        exception_class = import_string(self.exception_class_path)

        if not isclass(exception_class) or not issubclass(
            exception_class, BaseException
        ):
            raise ValueError(
                f"{self.exception_class_path!r} does not reference a valid exception."
            )

        return exception_class


@dataclass(frozen=True, slots=PY311, kw_only=True)  # type: ignore[literal-required]
class TaskResult(Generic[T]):
    task: Task
    """Task for which this is a result"""

    id: str
    """A unique identifier for the task result"""

    status: TaskResultStatus
    """Status of the running Task"""

    enqueued_at: datetime | None
    """Time this Task was enqueued"""

    started_at: datetime | None
    """Time this Task was started"""

    finished_at: datetime | None
    """Time this Task was finished"""

    last_attempted_at: datetime | None
    """Time this Task was last attempted to be run"""

    args: list
    """The arguments to pass to the task function"""

    kwargs: dict[str, Any]
    """The keyword arguments to pass to the task function"""

    backend: str
    """The name of the backend the task is run with"""

    errors: list[TaskError]
    """Errors raised when running the task"""

    worker_ids: list[str]
    """The workers which have processed the task"""

    _return_value: T | None = field(init=False, default=None)

    def __post_init__(self) -> None:
        object.__setattr__(self, "args", normalize_json(self.args))
        object.__setattr__(self, "kwargs", normalize_json(self.kwargs))

    @property
    def return_value(self) -> T | None:
        """
        The return value of the task.

        If the task didn't succeed, an exception is raised.
        This is to distinguish against the task returning None.
        """
        if self.status == TaskResultStatus.SUCCEEDED:
            return cast(T, self._return_value)
        elif self.status == TaskResultStatus.FAILED:
            raise ValueError("Task failed")
        else:
            raise ValueError("Task has not finished yet")

    @property
    def is_finished(self) -> bool:
        """Has the task finished?"""
        return self.status in {TaskResultStatus.FAILED, TaskResultStatus.SUCCEEDED}

    @property
    def attempts(self) -> int:
        return len(self.worker_ids)

    def refresh(self) -> None:
        """
        Reload the cached task data from the task store
        """
        refreshed_task = self.task.get_backend().get_result(self.id)

        for attr in TASK_REFRESH_ATTRS:
            object.__setattr__(self, attr, getattr(refreshed_task, attr))

    async def arefresh(self) -> None:
        """
        Reload the cached task data from the task store
        """
        refreshed_task = await self.task.get_backend().aget_result(self.id)

        for attr in TASK_REFRESH_ATTRS:
            object.__setattr__(self, attr, getattr(refreshed_task, attr))


@dataclass(frozen=True, slots=PY311, kw_only=True)  # type: ignore[literal-required]
class TaskContext:
    task_result: TaskResult

    @property
    def attempt(self) -> int:
        return self.task_result.attempts



# Generated by Django 5.0.6 on 2024-07-04 09:31

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "django_tasks_database",
            "0003_dbtaskresult_enqueued_at_dbtaskresult_finished_at",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="dbtaskresult",
            name="started_at",
            field=models.DateTimeField(null=True),
        ),
    ]



import logging
from argparse import ArgumentParser, ArgumentTypeError
from datetime import timedelta

from django.core.management.base import BaseCommand
from django.db.models import Q
from django.utils import timezone

from django_tasks import (
    DEFAULT_TASK_BACKEND_ALIAS,
    DEFAULT_TASK_QUEUE_NAME,
    task_backends,
)
from django_tasks.backends.database.backend import DatabaseBackend
from django_tasks.backends.database.models import DBTaskResult
from django_tasks.base import TaskResultStatus
from django_tasks.exceptions import InvalidTaskBackendError

logger = logging.getLogger("django_tasks.backends.database.prune_db_task_results")


def valid_backend_name(val: str) -> DatabaseBackend:
    try:
        backend = task_backends[val]
    except InvalidTaskBackendError as e:
        raise ArgumentTypeError(e.args[0]) from e
    if not isinstance(backend, DatabaseBackend):
        raise ArgumentTypeError(f"Backend '{val}' is not a database backend")
    return backend


def valid_positive_int(val: str) -> int:
    num = int(val)
    if num < 0:
        raise ArgumentTypeError("Must be greater than zero")
    return num


class Command(BaseCommand):
    help = "Prune finished database task results"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "--backend",
            nargs="?",
            default=DEFAULT_TASK_BACKEND_ALIAS,
            type=valid_backend_name,
            dest="backend",
            help="The backend to operate on (default: %(default)r)",
        )
        parser.add_argument(
            "--queue-name",
            nargs="?",
            default=DEFAULT_TASK_QUEUE_NAME,
            type=str,
            help="The queues to process. Separate multiple with a comma. To process all queues, use '*' (default: %(default)r)",
        )
        parser.add_argument(
            "--min-age-days",
            nargs="?",
            default=14,
            type=valid_positive_int,
            help="The minimum age (in days) of a finished task result to be pruned (default: %(default)r)",
        )
        parser.add_argument(
            "--failed-min-age-days",
            nargs="?",
            default=None,
            type=valid_positive_int,
            help="The minimum age (in days) of a failed task result to be pruned (default: min-age-days)",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Don't delete the task results, just show how many would be deleted",
        )

    def configure_logging(self, verbosity: int) -> None:
        if verbosity == 0:
            logger.setLevel(logging.WARNING)
        elif verbosity == 1:
            logger.setLevel(logging.INFO)
        else:
            logger.setLevel(logging.DEBUG)

        # If no handler is configured, the logs won't show,
        # regardless of the set level.
        if not logger.hasHandlers():
            logger.addHandler(logging.StreamHandler(self.stdout))

    def handle(
        self,
        *,
        verbosity: int,
        backend: DatabaseBackend,
        min_age_days: int,
        failed_min_age_days: int | None,
        queue_name: str,
        dry_run: bool,
        **options: dict,
    ) -> None:
        self.configure_logging(verbosity)

        min_age = timezone.now() - timedelta(days=min_age_days)
        failed_min_age = (
            (timezone.now() - timedelta(days=failed_min_age_days))
            if failed_min_age_days
            else None
        )

        results = DBTaskResult.objects.finished().filter(backend_name=backend.alias)

        queue_names = queue_name.split(",")
        if "*" not in queue_names:
            results = results.filter(queue_name__in=queue_names)

        if failed_min_age is None:
            results = results.filter(finished_at__lte=min_age)
        else:
            results = results.filter(
                Q(status=TaskResultStatus.SUCCEEDED, finished_at__lte=min_age)
                | Q(status=TaskResultStatus.FAILED, finished_at__lte=failed_min_age)
            )

        if dry_run:
            logger.info("Would delete %d task result(s)", results.count())
        else:
            deleted, _ = results.delete()
            logger.info("Deleted %d task result(s)", deleted)



from typing import Any

from django.db.models.signals import pre_save
from django.dispatch import receiver

from .models import DBTaskResult, get_date_max


@receiver(pre_save, sender=DBTaskResult)
def set_run_after(sender: Any, instance: DBTaskResult, **kwargs: Any) -> None:
    if instance.run_after is None:
        instance.run_after = get_date_max()



# Generated by Django 4.2.13 on 2024-11-21 13:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_tasks_database", "0009_remove_results_field"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dbtaskresult",
            name="status",
            field=models.CharField(
                choices=[
                    ("NEW", "New"),
                    ("RUNNING", "Running"),
                    ("FAILED", "Failed"),
                    ("SUCCEEDED", "Succeeded"),
                ],
                default="NEW",
                max_length=9,
                verbose_name="status",
            ),
        ),
    ]



from collections.abc import Iterable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, TypeVar

from django.apps import apps
from django.core import checks
from django.core.exceptions import ValidationError
from django.db import transaction
from django.utils.version import PY311
from typing_extensions import ParamSpec

from django_tasks.backends.base import BaseTaskBackend
from django_tasks.base import Task
from django_tasks.base import TaskResult as BaseTaskResult
from django_tasks.exceptions import TaskResultDoesNotExist
from django_tasks.signals import task_enqueued
from django_tasks.utils import normalize_json

if TYPE_CHECKING:
    from .models import DBTaskResult

T = TypeVar("T")
P = ParamSpec("P")


@dataclass(frozen=True, slots=PY311, kw_only=True)  # type: ignore[literal-required]
class TaskResult(BaseTaskResult[T]):
    db_result: "DBTaskResult"


class DatabaseBackend(BaseTaskBackend):
    supports_async_task = True
    supports_get_result = True
    supports_defer = True
    supports_priority = True

    def _task_to_db_task(
        self,
        task: Task[P, T],
        args: P.args,  # type:ignore[valid-type]
        kwargs: P.kwargs,  # type:ignore[valid-type]
    ) -> "DBTaskResult":
        from .models import DBTaskResult

        return DBTaskResult(
            args_kwargs=normalize_json({"args": args, "kwargs": kwargs}),
            priority=task.priority,
            task_path=task.module_path,
            queue_name=task.queue_name,
            run_after=task.run_after,  # type: ignore[misc]
            backend_name=self.alias,
        )

    def enqueue(
        self,
        task: Task[P, T],
        args: P.args,  # type:ignore[valid-type]
        kwargs: P.kwargs,  # type:ignore[valid-type]
    ) -> TaskResult[T]:
        self.validate_task(task)

        db_result = self._task_to_db_task(task, args, kwargs)

        def save_result() -> None:
            db_result.save()
            task_enqueued.send(type(self), task_result=db_result.task_result)

        if self._get_enqueue_on_commit_for_task(task):
            transaction.on_commit(save_result)
        else:
            save_result()

        return db_result.task_result

    def get_result(self, result_id: str) -> TaskResult:
        from .models import DBTaskResult

        try:
            return DBTaskResult.objects.get(id=result_id).task_result
        except (DBTaskResult.DoesNotExist, ValidationError) as e:
            raise TaskResultDoesNotExist(result_id) from e

    async def aget_result(self, result_id: str) -> TaskResult:
        from .models import DBTaskResult

        try:
            return (await DBTaskResult.objects.aget(id=result_id)).task_result
        except (DBTaskResult.DoesNotExist, ValidationError) as e:
            raise TaskResultDoesNotExist(result_id) from e

    def check(self, **kwargs: Any) -> Iterable[checks.CheckMessage]:
        yield from super().check(**kwargs)

        backend_name = self.__class__.__name__

        if not apps.is_installed("django_tasks.backends.database"):
            yield checks.Error(
                f"{backend_name} configured as django_tasks backend, but database app not installed",
                "Insert 'django_tasks.backends.database' in INSTALLED_APPS",
            )



import logging
from functools import partial
from typing import TypeVar

from django.db import transaction
from django.utils import timezone
from typing_extensions import ParamSpec

from django_tasks.base import Task, TaskContext, TaskError, TaskResult, TaskResultStatus
from django_tasks.signals import task_enqueued, task_finished, task_started
from django_tasks.utils import (
    get_exception_traceback,
    get_module_path,
    get_random_id,
    normalize_json,
)

from .base import BaseTaskBackend

logger = logging.getLogger(__name__)


T = TypeVar("T")
P = ParamSpec("P")


class ImmediateBackend(BaseTaskBackend):
    supports_async_task = True
    supports_priority = True

    def __init__(self, alias: str, params: dict):
        super().__init__(alias, params)

        self.worker_id = get_random_id()

    def _execute_task(self, task_result: TaskResult) -> None:
        """
        Execute the Task for the given TaskResult, mutating it with the outcome
        """
        object.__setattr__(task_result, "enqueued_at", timezone.now())
        task_enqueued.send(type(self), task_result=task_result)

        task = task_result.task

        task_start_time = timezone.now()

        object.__setattr__(task_result, "status", TaskResultStatus.RUNNING)
        object.__setattr__(task_result, "started_at", task_start_time)
        object.__setattr__(task_result, "last_attempted_at", task_start_time)
        task_result.worker_ids.append(self.worker_id)
        task_started.send(sender=type(self), task_result=task_result)

        try:
            if task.takes_context:
                raw_return_value = task.call(
                    TaskContext(task_result=task_result),
                    *task_result.args,
                    **task_result.kwargs,
                )
            else:
                raw_return_value = task.call(*task_result.args, **task_result.kwargs)

            object.__setattr__(
                task_result,
                "_return_value",
                normalize_json(raw_return_value),
            )
        except KeyboardInterrupt:
            # If the user tried to terminate, let them
            raise
        except BaseException as e:
            object.__setattr__(task_result, "finished_at", timezone.now())

            task_result.errors.append(
                TaskError(
                    exception_class_path=get_module_path(type(e)),
                    traceback=get_exception_traceback(e),
                )
            )

            object.__setattr__(task_result, "status", TaskResultStatus.FAILED)

            task_finished.send(type(self), task_result=task_result)
        else:
            object.__setattr__(task_result, "finished_at", timezone.now())
            object.__setattr__(task_result, "status", TaskResultStatus.SUCCEEDED)

            task_finished.send(type(self), task_result=task_result)

    def enqueue(
        self,
        task: Task[P, T],
        args: P.args,  # type:ignore[valid-type]
        kwargs: P.kwargs,  # type:ignore[valid-type]
    ) -> TaskResult[T]:
        self.validate_task(task)

        task_result = TaskResult[T](
            task=task,
            id=get_random_id(),
            status=TaskResultStatus.READY,
            enqueued_at=None,
            started_at=None,
            last_attempted_at=None,
            finished_at=None,
            args=args,
            kwargs=kwargs,
            backend=self.alias,
            errors=[],
            worker_ids=[],
        )

        if self._get_enqueue_on_commit_for_task(task) is not False:
            transaction.on_commit(partial(self._execute_task, task_result))
        else:
            self._execute_task(task_result)

        return task_result



from django.contrib import admin
from django.http import HttpRequest

from .models import DBTaskResult


@admin.register(DBTaskResult)
class DBTaskResultAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "task_name",
        "status",
        "enqueued_at",
        "started_at",
        "finished_at",
        "priority",
        "queue_name",
    )
    list_filter = ("status", "priority", "queue_name")
    ordering = ["-enqueued_at"]

    def has_add_permission(
        self, request: HttpRequest, obj: DBTaskResult | None = None
    ) -> bool:
        return False

    def has_delete_permission(
        self, request: HttpRequest, obj: DBTaskResult | None = None
    ) -> bool:
        return False

    def has_change_permission(
        self, request: HttpRequest, obj: DBTaskResult | None = None
    ) -> bool:
        return False

    def get_readonly_fields(
        self, request: HttpRequest, obj: DBTaskResult | None = None
    ) -> list[str]:
        return [f.name for f in self.model._meta.fields]



# Generated by Django 4.2.13 on 2024-06-08 15:24

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_tasks_database", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="dbtaskresult",
            options={
                "ordering": [
                    models.OrderBy(models.F("priority"), descending=True),
                    models.OrderBy(
                        models.F("run_after"), descending=True, nulls_last=True
                    ),
                ],
                "verbose_name": "Task Result",
                "verbose_name_plural": "Task Results",
            },
        ),
    ]



# ruff: noqa: E402
import django_stubs_ext

django_stubs_ext.monkeypatch()

import importlib.metadata

from django.conf import global_settings
from django.utils.connection import BaseConnectionHandler, ConnectionProxy
from django.utils.module_loading import import_string

from .backends.base import BaseTaskBackend
from .base import (
    DEFAULT_TASK_BACKEND_ALIAS,
    DEFAULT_TASK_QUEUE_NAME,
    TaskContext,
    TaskResult,
    TaskResultStatus,
    task,
)
from .exceptions import InvalidTaskBackendError

__version__ = importlib.metadata.version(__name__)

__all__ = [
    "task_backends",
    "default_task_backend",
    "DEFAULT_TASK_BACKEND_ALIAS",
    "DEFAULT_TASK_QUEUE_NAME",
    "TaskResultStatus",
    "TaskResult",
    "TaskContext",
    "task",
]


class TaskBackendHandler(BaseConnectionHandler[BaseTaskBackend]):
    settings_name = "TASKS"
    exception_class = InvalidTaskBackendError

    def configure_settings(self, settings: dict | None) -> dict:
        try:
            task_settings = super().configure_settings(settings)
        except AttributeError:
            # HACK: Force a default task backend.
            # Can be replaced with `django.conf.global_settings` once vendored.
            task_settings = None

        if task_settings is None or task_settings is getattr(
            global_settings, self.settings_name, None
        ):
            task_settings = {
                DEFAULT_TASK_BACKEND_ALIAS: {
                    "BACKEND": "django_tasks.backends.immediate.ImmediateBackend"
                }
            }

        return task_settings

    def create_connection(self, alias: str) -> BaseTaskBackend:
        params = self.settings[alias]

        backend = params["BACKEND"]

        try:
            backend_cls = import_string(backend)
        except ImportError as e:
            raise InvalidTaskBackendError(
                f"Could not find backend '{backend}': {e}"
            ) from e

        return backend_cls(alias=alias, params=params)  # type:ignore[no-any-return]


task_backends = TaskBackendHandler()

default_task_backend: BaseTaskBackend = ConnectionProxy(  # type:ignore[assignment]
    task_backends, DEFAULT_TASK_BACKEND_ALIAS
)



from django.apps import AppConfig


class TasksAppConfig(AppConfig):
    name = "django_tasks.backends.database"
    label = "django_tasks_database"
    verbose_name = "Tasks Database Backend"

    def ready(self) -> None:
        from . import signal_handlers  # noqa



from copy import deepcopy
from functools import partial
from typing import TypeVar

from django.db import transaction
from django.utils import timezone
from typing_extensions import ParamSpec

from django_tasks.base import Task, TaskResult, TaskResultStatus
from django_tasks.exceptions import TaskResultDoesNotExist
from django_tasks.signals import task_enqueued
from django_tasks.utils import get_random_id

from .base import BaseTaskBackend

T = TypeVar("T")
P = ParamSpec("P")


class DummyBackend(BaseTaskBackend):
    supports_defer = True
    supports_async_task = True
    supports_priority = True
    results: list[TaskResult]

    def __init__(self, alias: str, params: dict) -> None:
        super().__init__(alias, params)

        self.results = []

    def _store_result(self, result: TaskResult) -> None:
        object.__setattr__(result, "enqueued_at", timezone.now())
        self.results.append(result)
        task_enqueued.send(type(self), task_result=result)

    def enqueue(
        self,
        task: Task[P, T],
        args: P.args,  # type:ignore[valid-type]
        kwargs: P.kwargs,  # type:ignore[valid-type]
    ) -> TaskResult[T]:
        self.validate_task(task)

        result = TaskResult[T](
            task=task,
            id=get_random_id(),
            status=TaskResultStatus.READY,
            enqueued_at=None,
            started_at=None,
            last_attempted_at=None,
            finished_at=None,
            args=args,
            kwargs=kwargs,
            backend=self.alias,
            errors=[],
            worker_ids=[],
        )

        if self._get_enqueue_on_commit_for_task(task) is not False:
            transaction.on_commit(partial(self._store_result, result))
        else:
            self._store_result(result)

        # Copy the task to prevent mutation issues
        return deepcopy(result)

    # We don't set `supports_get_result` as the results are scoped to the current thread
    def get_result(self, result_id: str) -> TaskResult:
        try:
            return next(result for result in self.results if result.id == result_id)
        except StopIteration:
            raise TaskResultDoesNotExist(result_id) from None

    async def aget_result(self, result_id: str) -> TaskResult:
        try:
            return next(result for result in self.results if result.id == result_id)
        except StopIteration:
            raise TaskResultDoesNotExist(result_id) from None

    def clear(self) -> None:
        self.results.clear()



from collections.abc import Iterable
from importlib.metadata import version
from types import TracebackType
from typing import Any, TypeVar, cast

import django_rq
from django.apps import apps
from django.core.checks import messages
from django.core.exceptions import SuspiciousOperation
from django.db import transaction
from django.utils.functional import cached_property
from redis.client import Redis
from rq.defaults import UNSERIALIZABLE_RETURN_VALUE_PAYLOAD
from rq.exceptions import NoSuchJobError
from rq.job import Callback, JobStatus
from rq.job import Job as BaseJob
from rq.registry import ScheduledJobRegistry
from rq.results import Result
from typing_extensions import ParamSpec

from django_tasks.backends.base import BaseTaskBackend
from django_tasks.base import (
    TASK_DEFAULT_PRIORITY,
    TASK_MAX_PRIORITY,
    Task,
    TaskContext,
    TaskError,
    TaskResult,
    TaskResultStatus,
)
from django_tasks.compat import TASK_CLASSES
from django_tasks.exceptions import TaskResultDoesNotExist
from django_tasks.signals import task_enqueued, task_finished, task_started
from django_tasks.utils import get_module_path, get_random_id

T = TypeVar("T")
P = ParamSpec("P")

RQ_STATUS_TO_RESULT_STATUS = {
    JobStatus.QUEUED: TaskResultStatus.READY,
    JobStatus.FINISHED: TaskResultStatus.SUCCEEDED,
    JobStatus.FAILED: TaskResultStatus.FAILED,
    JobStatus.STARTED: TaskResultStatus.RUNNING,
    JobStatus.DEFERRED: TaskResultStatus.READY,
    JobStatus.SCHEDULED: TaskResultStatus.READY,
    JobStatus.STOPPED: TaskResultStatus.FAILED,
    JobStatus.CANCELED: TaskResultStatus.FAILED,
    None: TaskResultStatus.READY,
}


class Job(BaseJob):
    def perform(self) -> Any:
        task_started.send(
            type(self.task_result.task.get_backend()), task_result=self.task_result
        )

        return super().perform()

    def _execute(self) -> Any:
        """
        Shim RQ's `Job` to call the underlying `Task` function.
        """
        try:
            if self.func.takes_context:
                return self.func.call(
                    TaskContext(task_result=self.task_result), *self.args, **self.kwargs
                )
            return self.func.call(*self.args, **self.kwargs)
        finally:
            # Clear the task result cache, as it's changed now
            self.__dict__.pop("task_result", None)
            pass

    @property
    def func(self) -> Task:
        func = super().func

        if not isinstance(func, TASK_CLASSES):
            raise SuspiciousOperation(
                f"Task {self.id} does not point to a Task ({self.func_name})"
            )

        return func  # type: ignore[no-any-return]

    @cached_property
    def task_result(self) -> TaskResult:
        task: Task = self.func

        scheduled_job_registry = ScheduledJobRegistry(  # type: ignore[no-untyped-call]
            queue=django_rq.get_queue(self.origin)
        )

        if self.is_scheduled:
            run_after = scheduled_job_registry.get_scheduled_time(self)
        else:
            run_after = None

        task_result: TaskResult = TaskResult(
            task=task.using(
                priority=TASK_DEFAULT_PRIORITY,
                queue_name=self.origin,
                run_after=run_after,
                backend=self.meta["backend_name"],
            ),
            id=self.id,
            status=RQ_STATUS_TO_RESULT_STATUS[self.get_status()],
            enqueued_at=self.enqueued_at,
            started_at=self.started_at,
            last_attempted_at=self.started_at,
            finished_at=self.ended_at,
            args=list(self.args),
            kwargs=self.kwargs,
            backend=self.meta["backend_name"],
            errors=[],
            worker_ids=[],
        )

        exception_classes = self.meta.get("_django_tasks_exceptions", []).copy()

        rq_results = self.results()

        for rq_result in rq_results:
            task_result.worker_ids.append(rq_result.worker_name)
            if rq_result.type == Result.Type.FAILED:
                task_result.errors.append(
                    TaskError(
                        exception_class_path=(
                            exception_classes.pop()
                            if len(exception_classes) > 0
                            else get_module_path(Exception)
                        ),
                        traceback=rq_result.exc_string,  # type: ignore[arg-type]
                    )
                )

        if self.worker_name and task_result.status == TaskResultStatus.RUNNING:
            task_result.worker_ids.append(self.worker_name)

        if rq_results:
            object.__setattr__(
                task_result, "_return_value", rq_results[-1].return_value
            )
            object.__setattr__(
                task_result, "last_attempted_at", rq_results[-1].created_at
            )

        # If the return value couldn't be serialized, a specific string is saved instead.
        if task_result._return_value == UNSERIALIZABLE_RETURN_VALUE_PAYLOAD:
            # In these cases, the task should be marked as failed instead
            object.__setattr__(task_result, "status", TaskResultStatus.FAILED)

            task_result.errors.append(
                TaskError(
                    exception_class_path=get_module_path(Exception),
                    traceback=task_result._return_value,  # The traceback isn't stored, so this is the best we can do
                )
            )

            object.__setattr__(task_result, "_return_value", None)

        return task_result


def failed_callback(
    job: Job,
    connection: Redis | None,
    exception_class: type[Exception],
    exception_value: Exception,
    traceback: TracebackType,
) -> None:
    # Smuggle the exception class through meta
    job.meta.setdefault("_django_tasks_exceptions", []).append(
        get_module_path(exception_class)
    )
    job.save_meta()  # type: ignore[no-untyped-call]

    task_result = job.task_result

    object.__setattr__(task_result, "status", TaskResultStatus.FAILED)

    task_finished.send(type(task_result.task.get_backend()), task_result=task_result)


def success_callback(job: Job, connection: Redis | None, result: Any) -> None:
    task_result = job.task_result

    object.__setattr__(task_result, "status", TaskResultStatus.SUCCEEDED)

    task_finished.send(type(task_result.task.get_backend()), task_result=task_result)


class RQBackend(BaseTaskBackend):
    supports_async_task = True
    supports_get_result = True
    supports_defer = True

    def __init__(self, alias: str, params: dict) -> None:
        super().__init__(alias, params)

        if not self.queues:
            self.queues = set(django_rq.settings.QUEUES.keys())

    def enqueue(
        self,
        task: Task[P, T],
        args: P.args,  # type:ignore[valid-type]
        kwargs: P.kwargs,  # type:ignore[valid-type]
    ) -> TaskResult[T]:
        self.validate_task(task)

        task_result = TaskResult[T](
            task=task,
            id=get_random_id(),
            status=TaskResultStatus.READY,
            enqueued_at=None,
            started_at=None,
            last_attempted_at=None,
            finished_at=None,
            args=args,
            kwargs=kwargs,
            backend=self.alias,
            errors=[],
            worker_ids=[],
        )

        queue = django_rq.get_queue(task.queue_name, job_class=Job)

        job = queue.create_job(
            task.module_path,
            args=args,
            kwargs=kwargs,
            job_id=task_result.id,
            status=JobStatus.SCHEDULED if task.run_after else JobStatus.QUEUED,
            on_failure=Callback(failed_callback),
            on_success=Callback(success_callback),
            meta={"backend_name": self.alias},
        )

        def save_result() -> None:
            nonlocal job
            if task.run_after is None:
                job = queue.enqueue_job(
                    job, at_front=task.priority == TASK_MAX_PRIORITY
                )
            else:
                job = queue.schedule_job(job, task.run_after)

            object.__setattr__(task_result, "enqueued_at", job.enqueued_at)
            task_enqueued.send(type(self), task_result=task_result)

        if self._get_enqueue_on_commit_for_task(task):
            transaction.on_commit(save_result)
        else:
            save_result()

        return task_result

    def _get_queues(self) -> list[django_rq.queues.DjangoRQ]:
        return django_rq.queues.get_queues(*self.queues, job_class=Job)  # type: ignore[no-any-return,no-untyped-call]

    def _get_job(self, result_id: str) -> Job | None:
        try:
            return cast(
                Job, Job.fetch(result_id, connection=django_rq.get_connection())
            )
        except NoSuchJobError:
            return None

    def get_result(self, result_id: str) -> TaskResult:
        job = self._get_job(result_id)

        if job is None:
            raise TaskResultDoesNotExist(result_id)

        return job.task_result

    def check(self, **kwargs: Any) -> Iterable[messages.CheckMessage]:
        yield from super().check(**kwargs)

        backend_name = self.__class__.__name__

        if not apps.is_installed("django_rq"):
            yield messages.Error(
                f"{backend_name} configured as django_tasks backend, but django_rq app not installed",
                "Insert 'django_rq' in INSTALLED_APPS",
            )

        for queue_name in self.queues:
            try:
                django_rq.get_queue(queue_name)
            except KeyError:
                yield messages.Error(
                    f"{queue_name!r} is not configured for django-rq",
                    f"Add {queue_name!r} to RQ_QUEUES",
                )

        if tuple(map(int, (version("rq").split(".")))) < (2, 5, 0):
            yield messages.Error(
                "Only rq >= 2.5.0 is supported, found " + version("rq"),
                "Install a newer version of rq",
            )



# Generated by Django 4.2.13 on 2024-08-23 14:51

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("django_tasks_database", "0008_separate_results_field"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="dbtaskresult",
            name="result",
        ),
    ]



from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from inspect import iscoroutinefunction
from typing import Any, TypeVar

from asgiref.sync import sync_to_async
from django.conf import settings
from django.core import checks
from django.db import connections
from django.utils import timezone
from django.utils.inspect import get_func_args
from typing_extensions import ParamSpec

from django_tasks.base import (
    TASK_DEFAULT_PRIORITY,
    TASK_MAX_PRIORITY,
    TASK_MIN_PRIORITY,
    Task,
    TaskResult,
)
from django_tasks.exceptions import InvalidTaskError
from django_tasks.utils import is_module_level_function

T = TypeVar("T")
P = ParamSpec("P")


class BaseTaskBackend(metaclass=ABCMeta):
    alias: str
    enqueue_on_commit: bool

    task_class = Task

    supports_defer = False
    """Does the backend support Tasks to be enqueued with the run_after attribute?"""

    supports_async_task = False
    """Does the backend support coroutines to be enqueued?"""

    supports_get_result = False
    """Does the backend support results being retrieved (from any thread / process)?"""

    supports_priority = False
    """Does the backend support tasks being executed in a given priority order?"""

    def __init__(self, alias: str, params: dict) -> None:
        from django_tasks import DEFAULT_TASK_QUEUE_NAME

        self.alias = alias
        self.queues = set(params.get("QUEUES", [DEFAULT_TASK_QUEUE_NAME]))
        self.enqueue_on_commit = bool(params.get("ENQUEUE_ON_COMMIT", True))
        self.options = params.get("OPTIONS", {})

    def _get_enqueue_on_commit_for_task(self, task: Task) -> bool:
        """
        Determine the correct `enqueue_on_commit` setting to use for a given task.
        """

        # If the task defines it, use that, otherwise, fall back to the backend.
        return (
            task.enqueue_on_commit
            if task.enqueue_on_commit is not None
            else self.enqueue_on_commit
        )

    def validate_task(self, task: Task) -> None:
        """
        Determine whether the provided Task can be executed by the backend.
        """
        if not is_module_level_function(task.func):
            raise InvalidTaskError("Task function must be defined at a module level.")

        if not self.supports_async_task and iscoroutinefunction(task.func):
            raise InvalidTaskError("Backend does not support async tasks.")

        task_func_args = get_func_args(task.func)

        if task.takes_context and (
            not task_func_args or task_func_args[0] != "context"
        ):
            raise InvalidTaskError(
                "Task takes context but does not have a first argument of 'context'."
            )

        if not self.supports_priority and task.priority != TASK_DEFAULT_PRIORITY:
            raise InvalidTaskError(
                "Backend does not support setting priority of tasks."
            )

        if (
            task.priority < TASK_MIN_PRIORITY
            or task.priority > TASK_MAX_PRIORITY
            or int(task.priority) != task.priority
        ):
            raise InvalidTaskError(
                f"priority must be a whole number between {TASK_MIN_PRIORITY} and {TASK_MAX_PRIORITY}."
            )

        if not self.supports_defer and task.run_after is not None:
            raise InvalidTaskError("Backend does not support run_after.")

        if (
            settings.USE_TZ
            and task.run_after is not None
            and not timezone.is_aware(task.run_after)
        ):
            raise InvalidTaskError("run_after must be an aware datetime.")

        if self.queues and task.queue_name not in self.queues:
            raise InvalidTaskError(
                f"Queue '{task.queue_name}' is not valid for backend."
            )

    @abstractmethod
    def enqueue(
        self,
        task: Task[P, T],
        args: P.args,  # type:ignore[valid-type]
        kwargs: P.kwargs,  # type:ignore[valid-type]
    ) -> TaskResult[T]:
        """
        Queue up a task to be executed
        """

    async def aenqueue(
        self,
        task: Task[P, T],
        args: P.args,  # type:ignore[valid-type]
        kwargs: P.kwargs,  #  type:ignore[valid-type]
    ) -> TaskResult[T]:
        """
        Queue up a task function (or coroutine) to be executed
        """
        return await sync_to_async(self.enqueue, thread_sensitive=True)(
            task=task, args=args, kwargs=kwargs
        )

    def get_result(self, result_id: str) -> TaskResult:
        """
        Retrieve a result by id if it exists, otherwise raise
        ResultDoesNotExist.
        """
        raise NotImplementedError(
            "This backend does not support retrieving or refreshing results."
        )

    async def aget_result(self, result_id: str) -> TaskResult:
        """See get_result()."""
        return await sync_to_async(self.get_result, thread_sensitive=True)(
            result_id=result_id
        )

    def check(self, **kwargs: Any) -> Iterable[checks.CheckMessage]:
        if self.enqueue_on_commit and not connections._settings:  # type: ignore[attr-defined]
            yield checks.Error(
                "`ENQUEUE_ON_COMMIT` cannot be used when no databases are configured",
                hint="Set ENQUEUE_ON_COMMIT to False",
                id="tasks.E001",
            )

        if (
            self.enqueue_on_commit
            and not connections["default"].features.supports_transactions
        ):
            yield checks.Error(
                "ENQUEUE_ON_COMMIT cannot be used on a database which doesn't support transactions",
                hint="Set ENQUEUE_ON_COMMIT to False",
                id="tasks.E002",
            )



try:
    from django.tasks.base import Task as DjangoTask
except ImportError:
    DjangoTask = None

from .base import Task

__all__ = ["TASK_CLASSES"]

TASK_CLASSES = (Task, DjangoTask) if DjangoTask is not None else (Task,)



# Generated by Django 5.1.5 on 2025-04-27 12:00

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_tasks_database", "0014_remove_dbtaskresult_exception_data"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="dbtaskresult",
            options={
                "ordering": [
                    models.OrderBy(models.F("priority"), descending=True),
                    models.OrderBy(models.F("run_after"), nulls_last=True),
                ],
                "verbose_name": "Task Result",
                "verbose_name_plural": "Task Results",
            },
        ),
    ]



# Generated by Django 4.2.13 on 2024-08-23 14:38

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_tasks_database", "0006_alter_dbtaskresult_args_kwargs_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="dbtaskresult",
            name="exception_data",
            field=models.JSONField(
                default=None, null=True, verbose_name="exception data"
            ),
        ),
        migrations.AddField(
            model_name="dbtaskresult",
            name="return_value",
            field=models.JSONField(
                default=None, null=True, verbose_name="return value"
            ),
        ),
    ]



# Generated by Django 4.2.13 on 2024-08-23 14:38

from django.db import migrations, models
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps
from django.db.models.functions import Coalesce


def separate_exception_fields(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    DBTaskResult = apps.get_model("django_tasks_database", "DBTaskResult")

    DBTaskResult.objects.using(schema_editor.connection.alias).update(
        exception_class_path=Coalesce(
            models.F("exception_data__exc_type"), models.Value("", models.JSONField())
        ),
        traceback=Coalesce(
            models.F("exception_data__exc_traceback"),
            models.Value("", models.JSONField()),
        ),
    )


class Migration(migrations.Migration):
    dependencies = [
        ("django_tasks_database", "0012_add_separate_exception_fields"),
    ]

    operations = [migrations.RunPython(separate_exception_fields)]



# Generated by Django 5.2.2 on 2025-06-13 14:56

import datetime

from django.conf import settings
from django.db import migrations, models
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps


def update_status(apps: StateApps, schema_editor: BaseDatabaseSchemaEditor) -> None:
    DBTaskResult = apps.get_model("django_tasks_database", "DBTaskResult")

    DBTaskResult.objects.filter(status="NEW").update(status="READY")


def revert_status_rename(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    DBTaskResult = apps.get_model("django_tasks_database", "DBTaskResult")

    DBTaskResult.objects.filter(status="READY").update(status="NEW")


class Migration(migrations.Migration):
    dependencies = [
        ("django_tasks_database", "0015_correctly_order_run_after"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="dbtaskresult",
            options={
                "ordering": [
                    models.OrderBy(models.F("priority"), descending=True),
                    models.OrderBy(models.F("run_after")),
                ],
                "verbose_name": "Task Result",
                "verbose_name_plural": "Task Results",
            },
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="backend_name",
            field=models.CharField(max_length=32, verbose_name="backend name"),
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="queue_name",
            field=models.CharField(
                default="default", max_length=32, verbose_name="queue name"
            ),
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="run_after",
            field=models.DateTimeField(
                default=datetime.datetime(
                    9999,
                    1,
                    1,
                    0,
                    0,
                    tzinfo=datetime.timezone.utc if settings.USE_TZ else None,
                ),
                verbose_name="run after",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="status",
            field=models.CharField(
                choices=[
                    ("READY", "Ready"),
                    ("RUNNING", "Running"),
                    ("FAILED", "Failed"),
                    ("SUCCEEDED", "Succeeded"),
                ],
                default="READY",
                max_length=9,
                verbose_name="status",
            ),
        ),
        migrations.AddField(
            model_name="dbtaskresult",
            name="worker_ids",
            field=models.JSONField(default=list, verbose_name="worker id"),
        ),
        migrations.RunPython(update_status, revert_status_rename),
        migrations.AddIndex(
            model_name="dbtaskresult",
            index=models.Index(
                models.F("status"),
                models.OrderBy(models.F("priority"), descending=True),
                models.OrderBy(models.F("run_after")),
                condition=models.Q(("status", "READY")),
                name="django_task_new_ordering_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="dbtaskresult",
            index=models.Index(
                fields=["queue_name"], name="django_task_queue_n_99c321_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="dbtaskresult",
            index=models.Index(
                fields=["backend_name"], name="django_task_backend_071754_idx"
            ),
        ),
    ]



import logging
import math
import os
import random
import signal
import sys
import time
from argparse import ArgumentParser, ArgumentTypeError, BooleanOptionalAction
from types import FrameType

from django.conf import settings
from django.core.exceptions import SuspiciousOperation
from django.core.management.base import BaseCommand
from django.db import close_old_connections
from django.db.utils import OperationalError
from django.utils.autoreload import DJANGO_AUTORELOAD_ENV, run_with_reloader

from django_tasks import DEFAULT_TASK_BACKEND_ALIAS, task_backends
from django_tasks.backends.database.backend import DatabaseBackend
from django_tasks.backends.database.models import DBTaskResult
from django_tasks.backends.database.utils import exclusive_transaction
from django_tasks.base import DEFAULT_TASK_QUEUE_NAME, TaskContext
from django_tasks.exceptions import InvalidTaskBackendError
from django_tasks.signals import task_finished, task_started
from django_tasks.utils import get_random_id

package_logger = logging.getLogger("django_tasks")
logger = logging.getLogger("django_tasks.backends.database.db_worker")


class Worker:
    def __init__(
        self,
        *,
        queue_names: list[str],
        interval: float,
        batch: bool,
        backend_name: str,
        startup_delay: bool,
        max_tasks: int | None,
        worker_id: str,
    ):
        self.queue_names = queue_names
        self.process_all_queues = "*" in queue_names
        self.interval = interval
        self.batch = batch
        self.backend_name = backend_name
        self.startup_delay = startup_delay
        self.max_tasks = max_tasks

        self.running = True
        self.running_task = False
        self._run_tasks = 0

        self.worker_id = worker_id

    def shutdown(self, signum: int, frame: FrameType | None) -> None:
        if not self.running:
            logger.warning(
                "Received %s - terminating current task.", signal.strsignal(signum)
            )
            self.reset_signals()
            sys.exit(1)

        logger.warning(
            "Received %s - shutting down gracefully... (press Ctrl+C again to force)",
            signal.strsignal(signum),
        )
        self.running = False

        if not self.running_task:
            # If we're not currently running a task, exit immediately.
            # This is useful if we're currently in a `sleep`.
            sys.exit(0)

    def configure_signals(self) -> None:
        signal.signal(signal.SIGINT, self.shutdown)
        signal.signal(signal.SIGTERM, self.shutdown)
        if hasattr(signal, "SIGQUIT"):
            signal.signal(signal.SIGQUIT, self.shutdown)

    def reset_signals(self) -> None:
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        if hasattr(signal, "SIGQUIT"):
            signal.signal(signal.SIGQUIT, signal.SIG_DFL)

    def run(self) -> None:
        logger.info(
            "Starting worker worker_id=%s queues=%s",
            self.worker_id,
            ",".join(self.queue_names),
        )

        if self.startup_delay and self.interval:
            # Add a random small delay before starting to avoid a thundering herd
            time.sleep(random.random())  # noqa: S311

        while self.running:
            tasks = DBTaskResult.objects.ready().filter(backend_name=self.backend_name)
            if not self.process_all_queues:
                tasks = tasks.filter(queue_name__in=self.queue_names)

            # During this transaction, all "ready" tasks are locked. Therefore, it's important
            # it be as efficient as possible.
            with exclusive_transaction(tasks.db):
                try:
                    task_result = tasks.get_locked()
                except OperationalError as e:
                    # Ignore locked databases and keep trying.
                    # It should unlock eventually.
                    if "is locked" in e.args[0]:
                        task_result = None
                    else:
                        raise

                if task_result is not None:
                    # "claim" the task, so it isn't run by another worker process
                    task_result.claim(self.worker_id)

            if task_result is not None:
                self.run_task(task_result)

            if self.batch and task_result is None:
                # If we're running in "batch" mode, terminate the loop (and thus the worker)
                logger.info(
                    "No more tasks to run for worker_id=%s - exiting gracefully.",
                    self.worker_id,
                )
                return None

            if self.max_tasks is not None and self._run_tasks >= self.max_tasks:
                logger.info(
                    "Run maximum tasks (%d) on worker=%s - exiting gracefully.",
                    self._run_tasks,
                    self.worker_id,
                )
                return None

            # Emulate Django's request behaviour and check for expired
            # database connections periodically.
            close_old_connections()

            # If ctrl-c has just interrupted a task, self.running was cleared,
            # and we should not sleep, but rather exit immediately.
            if self.running and not task_result:
                # Wait before checking for another task
                time.sleep(self.interval)

    def run_task(self, db_task_result: DBTaskResult) -> None:
        """
        Run the given task, marking it as succeeded or failed.
        """
        try:
            self.running_task = True
            task = db_task_result.task
            task_result = db_task_result.task_result

            backend_type = task.get_backend()

            task_started.send(sender=backend_type, task_result=task_result)
            if task.takes_context:
                return_value = task.call(
                    TaskContext(task_result=task_result),
                    *task_result.args,
                    **task_result.kwargs,
                )
            else:
                return_value = task.call(*task_result.args, **task_result.kwargs)

            # Setting the return and success value inside the error handling,
            # So errors setting it (eg JSON encode) can still be recorded
            db_task_result.set_succeeded(return_value)
            task_finished.send(
                sender=backend_type, task_result=db_task_result.task_result
            )
        except BaseException as e:
            db_task_result.set_failed(e)
            try:
                sender = type(db_task_result.task.get_backend())
                task_result = db_task_result.task_result
            except (ModuleNotFoundError, SuspiciousOperation):
                logger.exception("Task id=%s failed unexpectedly", db_task_result.id)
            else:
                task_finished.send(
                    sender=sender,
                    task_result=task_result,
                )
        finally:
            self.running_task = False
            self._run_tasks += 1


def valid_backend_name(val: str) -> str:
    try:
        backend = task_backends[val]
    except InvalidTaskBackendError as e:
        raise ArgumentTypeError(e.args[0]) from e
    if not isinstance(backend, DatabaseBackend):
        raise ArgumentTypeError(f"Backend '{val}' is not a database backend")
    return val


def valid_interval(val: str) -> float:
    num = float(val)
    if not math.isfinite(num):
        raise ArgumentTypeError("Must be a finite floating point value")
    if num < 0:
        raise ArgumentTypeError("Must be greater than zero")
    return num


def valid_max_tasks(val: str) -> int:
    num = int(val)
    if num < 0:
        raise ArgumentTypeError("Must be greater than zero")
    return num


def validate_worker_id(val: str) -> str:
    if not val:
        raise ArgumentTypeError("Worker id must not be empty")
    if len(val) > 64:
        raise ArgumentTypeError("Worker ids must be shorter than 64 characters")
    return val


class Command(BaseCommand):
    help = "Run a database background worker"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "--queue-name",
            nargs="?",
            default=DEFAULT_TASK_QUEUE_NAME,
            type=str,
            help="The queues to process. Separate multiple with a comma. To process all queues, use '*' (default: %(default)r)",
        )
        parser.add_argument(
            "--interval",
            nargs="?",
            default=1,
            type=valid_interval,
            help="The interval (in seconds) to wait, when there are no tasks in the queue, before checking for tasks again (default: %(default)r)",
        )
        parser.add_argument(
            "--batch",
            action="store_true",
            help="Process all outstanding tasks, then exit. Can be used in combination with --max-tasks.",
        )
        parser.add_argument(
            "--reload",
            action=BooleanOptionalAction,
            default=settings.DEBUG,
            help="Reload the worker on code changes. Not recommended for production as tasks may not be stopped cleanly (default: DEBUG)",
        )
        parser.add_argument(
            "--backend",
            nargs="?",
            default=DEFAULT_TASK_BACKEND_ALIAS,
            type=valid_backend_name,
            dest="backend_name",
            help="The backend to operate on (default: %(default)r)",
        )
        parser.add_argument(
            "--no-startup-delay",
            action="store_false",
            dest="startup_delay",
            help="Don't add a small delay at startup.",
        )
        parser.add_argument(
            "--max-tasks",
            nargs="?",
            default=None,
            type=valid_max_tasks,
            help="If provided, the maximum number of tasks the worker will execute before exiting.",
        )
        parser.add_argument(
            "--worker-id",
            nargs="?",
            type=validate_worker_id,
            help="Worker id. MUST be unique across worker pool (default: auto-generate)",
            default=get_random_id(),
        )

    def configure_logging(self, verbosity: int) -> None:
        if verbosity == 0:
            package_logger.setLevel(logging.CRITICAL)
        elif verbosity == 1:
            package_logger.setLevel(logging.INFO)
        else:
            package_logger.setLevel(logging.DEBUG)

        # If no handler is configured, the logs won't show,
        # regardless of the set level.
        if not package_logger.hasHandlers():
            package_logger.addHandler(logging.StreamHandler(self.stdout))

    def handle(
        self,
        *,
        verbosity: int,
        queue_name: str,
        interval: float,
        batch: bool,
        backend_name: str,
        startup_delay: bool,
        reload: bool,
        max_tasks: int | None,
        worker_id: str,
        **options: dict,
    ) -> None:
        self.configure_logging(verbosity)

        if reload and batch:
            logger.warning(
                "Warning: --reload and --batch cannot be specified together. Disabling autoreload."
            )
            reload = False

        worker = Worker(
            queue_names=queue_name.split(","),
            interval=interval,
            batch=batch,
            backend_name=backend_name,
            startup_delay=startup_delay,
            max_tasks=max_tasks,
            worker_id=worker_id,
        )

        if reload:
            if os.environ.get(DJANGO_AUTORELOAD_ENV) == "true":
                # Only the child process should configure its signals
                worker.configure_signals()

            run_with_reloader(worker.run)
        else:
            worker.configure_signals()
            worker.run()



import logging

from asgiref.local import Local
from django.core.signals import setting_changed
from django.dispatch import Signal, receiver

from django_tasks import BaseTaskBackend, TaskResult, TaskResultStatus

task_enqueued = Signal()
task_finished = Signal()
task_started = Signal()


logger = logging.getLogger("django_tasks")


@receiver(setting_changed)
def clear_tasks_handlers(*, setting: str, **kwargs: dict) -> None:
    """
    Reset the connection handler whenever the settings change.
    """
    if setting == "TASKS":
        from django_tasks import task_backends

        task_backends._settings = task_backends.settings = (  # type:ignore[attr-defined]
            task_backends.configure_settings(None)
        )
        task_backends._connections = Local()  # type:ignore[attr-defined]


@receiver(task_enqueued)
def log_task_enqueued(
    sender: type[BaseTaskBackend], task_result: TaskResult, **kwargs: dict
) -> None:
    logger.debug(
        "Task id=%s path=%s enqueued backend=%s",
        task_result.id,
        task_result.task.module_path,
        task_result.backend,
    )


@receiver(task_started)
def log_task_started(
    sender: type[BaseTaskBackend], task_result: TaskResult, **kwargs: dict
) -> None:
    logger.info(
        "Task id=%s path=%s state=%s",
        task_result.id,
        task_result.task.module_path,
        task_result.status,
    )


@receiver(task_finished)
def log_task_finished(
    sender: type[BaseTaskBackend], task_result: TaskResult, **kwargs: dict
) -> None:
    if task_result.status == TaskResultStatus.FAILED:
        # Use exception to integrate with error monitoring tools
        log_method = logger.exception
    else:
        log_method = logger.info

    log_method(
        "Task id=%s path=%s state=%s",
        task_result.id,
        task_result.task.module_path,
        task_result.status,
    )



from collections.abc import Generator
from contextlib import contextmanager
from typing import Any
from uuid import UUID

import django
from django.db import transaction
from django.db.backends.base.base import BaseDatabaseWrapper


def connection_requires_manual_exclusive_transaction(
    connection: BaseDatabaseWrapper,
) -> bool:
    """
    Determine whether the backend requires manual transaction handling.

    Extracted from `exclusive_transaction` for unit testing purposes.
    """
    if connection.vendor != "sqlite":
        return False

    if django.VERSION < (5, 1):
        return True

    if not hasattr(connection, "transaction_mode"):
        # Manually called to set `transaction_mode`
        connection.get_connection_params()

    return connection.transaction_mode != "EXCLUSIVE"  # type:ignore[attr-defined,no-any-return]


@contextmanager
def exclusive_transaction(using: str | None = None) -> Generator[Any, Any, Any]:
    """
    Wrapper around `transaction.atomic` which ensures transactions on SQLite are exclusive.

    This functionality is built-in to Django 5.1+.
    """
    connection: BaseDatabaseWrapper = transaction.get_connection(using)

    if connection_requires_manual_exclusive_transaction(connection):
        with connection.cursor() as c:
            c.execute("BEGIN EXCLUSIVE")
            try:
                yield
            finally:
                c.execute("COMMIT")
    else:
        with transaction.atomic(using=using):
            yield


def normalize_uuid(val: str | UUID) -> str:
    """
    Normalize a UUID into its dashed representation.

    This works around engines like MySQL which don't store values in a uuid field,
    and thus drops the dashes.
    """
    if isinstance(val, str):
        val = UUID(val)

    return str(val)



# Generated by Django 4.2.14 on 2024-08-02 13:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_tasks_database", "0005_alter_dbtaskresult_priority_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dbtaskresult",
            name="args_kwargs",
            field=models.JSONField(verbose_name="args kwargs"),
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="backend_name",
            field=models.TextField(verbose_name="backend name"),
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="enqueued_at",
            field=models.DateTimeField(auto_now_add=True, verbose_name="enqueued at"),
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="finished_at",
            field=models.DateTimeField(null=True, verbose_name="finished at"),
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="priority",
            field=models.IntegerField(default=0, verbose_name="priority"),
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="queue_name",
            field=models.TextField(default="default", verbose_name="queue name"),
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="result",
            field=models.JSONField(default=None, null=True, verbose_name="result"),
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="run_after",
            field=models.DateTimeField(null=True, verbose_name="run after"),
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="started_at",
            field=models.DateTimeField(null=True, verbose_name="started at"),
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="status",
            field=models.CharField(
                choices=[
                    ("NEW", "New"),
                    ("RUNNING", "Running"),
                    ("FAILED", "Failed"),
                    ("COMPLETE", "Complete"),
                ],
                default="NEW",
                max_length=8,
                verbose_name="status",
            ),
        ),
        migrations.AlterField(
            model_name="dbtaskresult",
            name="task_path",
            field=models.TextField(verbose_name="task path"),
        ),
    ]



from collections.abc import Iterable, Sequence
from typing import Any

from django.apps.config import AppConfig
from django.core import checks

from django_tasks import task_backends


@checks.register
def check_tasks(
    app_configs: Sequence[AppConfig] = None, **kwargs: Any
) -> Iterable[checks.CheckMessage]:
    """Checks all registered task backends."""

    for backend in task_backends.all():
        yield from backend.check()



# Generated by Django 4.2.13 on 2024-08-23 14:38

from django.db import migrations, models
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps

from django_tasks import TaskResultStatus


def separate_results_field(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    DBTaskResult = apps.get_model("django_tasks_database", "DBTaskResult")

    # If a task succeeded, the result is its return value
    DBTaskResult.objects.using(schema_editor.connection.alias).filter(
        status=TaskResultStatus.SUCCEEDED
    ).update(return_value=models.F("result"))

    # If a task failed, the result is the exception data (or nothing)
    DBTaskResult.objects.using(schema_editor.connection.alias).filter(
        status=TaskResultStatus.FAILED
    ).update(exception_data=models.F("result"))


def merge_results_field(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    DBTaskResult = apps.get_model("django_tasks_database", "DBTaskResult")

    # If a task succeeded, the result is its return value
    DBTaskResult.objects.using(schema_editor.connection.alias).filter(
        status=TaskResultStatus.SUCCEEDED
    ).update(result=models.F("return_value"))

    # If a task failed, the result is the exception data (or nothing)
    DBTaskResult.objects.using(schema_editor.connection.alias).filter(
        status=TaskResultStatus.FAILED
    ).update(result=models.F("exception_data"))


class Migration(migrations.Migration):
    dependencies = [
        ("django_tasks_database", "0007_add_separate_results_fields"),
    ]

    operations = [migrations.RunPython(separate_results_field, merge_results_field)]



import inspect
import time
from collections.abc import Callable, Mapping, Sequence
from functools import wraps
from traceback import format_exception
from typing import Any, TypeVar

from django.utils.crypto import get_random_string
from typing_extensions import ParamSpec

T = TypeVar("T")
P = ParamSpec("P")


def is_module_level_function(func: Callable) -> bool:
    if not inspect.isfunction(func) or inspect.isbuiltin(func):
        return False

    if "<locals>" in func.__qualname__:
        return False

    return True


def normalize_json(obj: Any) -> Any:
    """Recursively normalize an object into JSON-compatible types."""
    match obj:
        case Mapping():
            return {normalize_json(k): normalize_json(v) for k, v in obj.items()}
        case bytes():
            try:
                return obj.decode("utf-8")
            except UnicodeDecodeError as e:
                raise ValueError(f"Unsupported value: {type(obj)}") from e
        case str() | int() | float() | bool() | None:
            return obj
        case Sequence():  # str and bytes were already handled.
            return [normalize_json(v) for v in obj]
        case _:  # Other types can't be serialized to JSON
            raise TypeError(f"Unsupported type: {type(obj)}")


def retry(*, retries: int = 3, backoff_delay: float = 0.1) -> Callable:
    """
    Retry the given code `retries` times, raising the final error.

    `backoff_delay` can be used to add a delay between attempts.
    """

    def wrapper(f: Callable[P, T]) -> Callable[P, T]:
        @wraps(f)
        def inner_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:  # type:ignore[return]
            for attempt in range(1, retries + 1):
                try:
                    return f(*args, **kwargs)
                except KeyboardInterrupt:
                    # Let the user ctrl-C out of the program without a retry
                    raise
                except BaseException:
                    if attempt == retries:
                        raise
                    time.sleep(backoff_delay)

        return inner_wrapper

    return wrapper


def get_module_path(val: Any) -> str:
    return f"{val.__module__}.{val.__qualname__}"


def get_exception_traceback(exc: BaseException) -> str:
    return "".join(format_exception(exc))


def get_random_id() -> str:
    """
    Return a random string for use as a Task or worker id.

    Whilst 64 characters is the max, just use 32 as a sensible middle-ground.
    """
    return get_random_string(32)
